/* File:  DynaDebug.cpp 

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is the DynaDebug class/library

The Initial Developer of the Original Code is Gregory Riggs.
Portions created by Gregory Riggs are Copyright (C) 1996, 2000, 2001
Gregory Riggs.  All Rights Reserved.

Contributor(s):
  Gregory Riggs <griggs@itqlata.com> <greg@nmia.com>
  ITQ LATA, LLC. <wctpxml@itqlata.com>

*/

#ifdef DYNADEBUG

#include "DynaDebug.hpp"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>

#ifdef _REENTRANT

pthread_once_t DynaDebug::debug_init_once = PTHREAD_ONCE_INIT;
pthread_key_t DynaDebug::debug_data_key;
pthread_mutex_t DynaDebug::debug_mutex;
int DynaDebug::thread_count = 0;

#else

bool DynaDebug::first_time = true;
DynaDebug::DebugContext DynaDebug::single_context;

#endif

bool DynaDebug::init_okay = false;
ofstream DynaDebug::derr;

//===========================================================================
// DynaDebug class
//===========================================================================

//---------------------------------------------------------------------------
// DynaDebug::DynaDebug(const char *funcname, int code)
//---------------------------------------------------------------------------

DynaDebug::DynaDebug(const char *funcname, int code)
    : curname(funcname)
{

#ifdef _REENTRANT

    pthread_once(&debug_init_once, DynaDebug::Initialize);

    context = (DebugContext *) pthread_getspecific(debug_data_key);
    if(context == NULL) 
    {
        context = new DebugContext;
        pthread_setspecific(debug_data_key, (void *)context);
        context->indent_flag = true;
        context->cur_indent = 0;

        pthread_mutex_lock(&debug_mutex);
        context->thread_number = thread_count++;
        pthread_mutex_unlock(&debug_mutex);
    }

#else

    context = &DynaDebug::single_context;

    if(first_time)
    {
        Initialize();
        first_time = false;
        context->indent_flag = true;
        context->cur_indent = 0;
    }

#endif

    context->sbuf = "";
    debug_flag = IsDebuggingOn(code);

    if(debug_flag) 
    {
        if(!context->indent_flag)
        {
            *this << endl;
        }
        indent();
        context->sbuf += "--> ";
        context->sbuf += curname;
        *this << endl;
        context->cur_indent += indent_size;
    }
}

//---------------------------------------------------------------------------
// DynaDebug::~DynaDebug()
//---------------------------------------------------------------------------

DynaDebug::~DynaDebug()
{
    if(context != NULL && debug_flag)
    {
        context->cur_indent -= indent_size;
        if(!context->indent_flag)
        {
            *this << endl;
        }
        indent();
        context->sbuf += "<-- ";
        context->sbuf += curname;
        *this << endl;
    }
}

//---------------------------------------------------------------------------
// void DynaDebug::Initialize(void)
//---------------------------------------------------------------------------

void DynaDebug::Initialize(void)
{

#ifdef _REENTRANT

    pthread_key_create(&debug_data_key, FreeContext);
    pthread_mutex_init(&debug_mutex, NULL);

#endif

    for(int i=0; i<num_labels; i++) 
    {
        cache[i] = false; 
    }

    if(environment_flagname != NULL && config_filename != NULL) 
    {
        char *flag = getenv(environment_flagname);
        if(flag != NULL && atoi(flag) && output_filename != NULL) 
        {
            derr.open(output_filename);
            if(derr)
            {
                derr << "** Dynamic Debugging Log file **" << endl;
                derr << flush;
                init_okay = true;
            }
        } 
    }

    return;
}

//---------------------------------------------------------------------------
// DynaDebug::FreeContext
//---------------------------------------------------------------------------

void DynaDebug::FreeContext(void *data)
{
    DebugContext *context;
    context = (DebugContext *)data;
    delete context;
}

//---------------------------------------------------------------------------
// int DynaDebug::IsDebuggingOn(int code)
//---------------------------------------------------------------------------

bool DynaDebug::IsDebuggingOn(int code)
{
    bool result;
    result = false;
    if(init_okay && code < num_labels && derr)
    {

#ifdef _REENTRANT
        pthread_mutex_lock(&debug_mutex);
#endif

        DynaDebug::UpdateCache();

#ifdef _REENTRANT
        pthread_mutex_unlock(&debug_mutex);
#endif

        result = cache[code];
    }
    return result;
}

//---------------------------------------------------------------------------
// void DynaDebug::UpdateCache(void)
//---------------------------------------------------------------------------

void DynaDebug::UpdateCache(void)
{
    static time_t last_mtime = 0;
    static struct stat stat_buf;

    if(stat(config_filename,&stat_buf)==0)
    {
        if(stat_buf.st_mtime > last_mtime)
        {
            last_mtime = stat_buf.st_mtime; 
            DynaDebug::LoadCache();
        }    
    } 
    else 
    {
        derr << "** Can't stat the file \"" 
             << config_filename << "\" **" << endl; 
        derr << flush;
    }
}

//---------------------------------------------------------------------------
// DynaDebug::LoadCache
//---------------------------------------------------------------------------

void DynaDebug::LoadCache(void)
{
    static char line_buf[MAX_LABEL_LENGTH];

    ifstream ifile(config_filename);
    if(!ifile)
    { 
        derr << "** Error reading file \"" << config_filename 
             << "\" **" << endl; 
        derr << flush;
        return;
    }

    derr << "** Reading in DYNADEBUG config file **" << endl;
    derr << flush;

    for(int i=0; i< num_labels; i++)
    {
        cache[i] = false;
    }

    while(ifile.getline(line_buf,MAX_LABEL_LENGTH)) 
    {
        for(int j=0; j<num_labels; j++)
        {
            int t = 0;
            while(line_buf[t] == ' ' || line_buf[t] == '\t')
            {
                t++;
            }
            if(strncmp(&line_buf[t], labels[j], strlen(labels[j])) == 0)
            {
                cache[j] = true;
            }
        }
    }

    return;
}

//---------------------------------------------------------------------------
// void DynaDebug::indent(void)
//---------------------------------------------------------------------------

void DynaDebug::indent(void)
{
    if(context->indent_flag) 
    {

#ifdef _REENTRANT
        snprintf(context->cbuf, 
                 DebugContext::CBUF_SIZE, "(%5.5d) ", 
                 context->thread_number);
        context->sbuf += context->cbuf;
#else
        snprintf(context->cbuf, 
                 DebugContext::CBUF_SIZE, "(%5.5d) ", 
                 (int)getpid());
        context->sbuf += context->cbuf;
#endif

        int i;
        for(i=0; i<context->cur_indent; i++)
        {
            context->sbuf += ' ';
        }
        context->indent_flag = false;
    }
}

//---------------------------------------------------------------------------
// DynaDebug& endl(DynaDebug& os)
//---------------------------------------------------------------------------

DynaDebug& endl(DynaDebug& os)
{ 
    os.context->sbuf += "\n";
    DynaDebug::derr << os.context->sbuf;
    DynaDebug::derr << flush;
    os.context->sbuf = "";
    os.context->indent_flag = true;
    return(os);
}

//---------------------------------------------------------------------------
// char
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const char c)
{
    if(debug_flag) 
    {
        indent(); 
        context->sbuf += c;
    } 
    return (*this);
}

//---------------------------------------------------------------------------
// signed char
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const signed char c)
{
    return this->operator<<((char)c);
}

//---------------------------------------------------------------------------
// unsigned char
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const unsigned char c)
{
    return this->operator<<((char)c);
}

//---------------------------------------------------------------------------
// char *
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const char * s)
{
    if(debug_flag) 
    {
        indent(); 
        if(s)
        {
            context->sbuf += s;
        }
        else 
        {
            context->sbuf += "(null)";
        }
    } 
    return (*this);
}

//---------------------------------------------------------------------------
// signed char *
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const signed char * s)
{
    return this->operator<<((char *)s);
}

//---------------------------------------------------------------------------
// unsigned char *
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const unsigned char * s)
{
    return this->operator<<((char *)s);
}

//---------------------------------------------------------------------------
// XString
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const XString & s)
{
    if(debug_flag)
    {
        indent(); 
        context->sbuf += s;
    }
    return (*this);
}

//---------------------------------------------------------------------------
// signed short
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const signed short i)
{
    if(debug_flag) 
    {
        indent(); 
        snprintf(context->cbuf, DebugContext::CBUF_SIZE, "%hi", i);
        context->sbuf += context->cbuf;
    } 
    return (*this);
}

//---------------------------------------------------------------------------
// unsigned short
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const unsigned short i)
{
    if(debug_flag) 
    {
        indent(); 
        snprintf(context->cbuf, DebugContext::CBUF_SIZE, "%hu", i);
        context->sbuf += context->cbuf;
    } 
    return (*this);
}

//---------------------------------------------------------------------------
// signed int
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const signed int i)
{
    if(debug_flag) 
    {
        indent(); 
        snprintf(context->cbuf, DebugContext::CBUF_SIZE, "%i", i);
        context->sbuf += context->cbuf;
    } 
    return (*this);
}

//---------------------------------------------------------------------------
// unsigned int
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const unsigned int i)
{
    if(debug_flag) 
    {
        indent(); 
        snprintf(context->cbuf, DebugContext::CBUF_SIZE, "%u", i);
        context->sbuf += context->cbuf;
    } 
    return (*this);
}

//---------------------------------------------------------------------------
// signed long
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const signed long i)
{
    if(debug_flag)
    {
        indent(); 
        snprintf(context->cbuf, DebugContext::CBUF_SIZE, "%li", i);
        context->sbuf += context->cbuf;
    } 
    return (*this);
}

//---------------------------------------------------------------------------
// unsigned long
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const unsigned long i)
{
    if(debug_flag) 
    {
        indent(); 
        snprintf(context->cbuf, DebugContext::CBUF_SIZE, "%lu", i);
        context->sbuf += context->cbuf;
    } 
    return (*this);
}

//---------------------------------------------------------------------------
// float
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const float f)
{
    if(debug_flag) 
    {
        indent(); 
        snprintf(context->cbuf, DebugContext::CBUF_SIZE, "%g", (double)f);
        context->sbuf += context->cbuf;
    } 
    return (*this);
}

//---------------------------------------------------------------------------
// double
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const double d)
{
    if(debug_flag) 
    {
        indent(); 
        snprintf(context->cbuf, DebugContext::CBUF_SIZE, "%g", d);
        context->sbuf += context->cbuf;
    } 
    return (*this);
}

//---------------------------------------------------------------------------
// long double
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(const long double d)
{
    return this->operator<<((double) d);
}

//---------------------------------------------------------------------------
// function name
//---------------------------------------------------------------------------

const char * DynaDebug::function_name(void)
{
    return curname;
}

//---------------------------------------------------------------------------
// DynaDebugFunc
//---------------------------------------------------------------------------

DynaDebug & DynaDebug::operator<<(DynaDebugFunc f)
{
    if(debug_flag)
    {
        return f(*this); 
    }
    else 
    {
        return (*this);
    }
}

#endif
