/* File:  DynaDebug.cpp 

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is the DynaDebug class/library

The Initial Developer of the Original Code is Gregory Riggs.
Portions created by Gregory Riggs are Copyright (C) 1996, 2000, 2001
Gregory Riggs.  All Rights Reserved.

Contributor(s):
  Gregory Riggs <griggs@itqlata.com> <greg@nmia.com>
  ITQ LATA, LLC. <wctpxml@itqlata.com>

*/

#ifndef DYNADEBUG_HPP
#define DYNADEBUG_HPP

#ifndef DYNADEBUG

#define DOBJ(X,Y)
#define DERR(X) 
#define ASSERT(X)

#else

#include <fstream.h>
#include "XString.hpp"

#ifdef _REENTRANT
#include <pthread.h>
#endif

// Note that the macros require semicolons after them

#define DOBJ(X,Y) DynaDebug derr(X,Y)
#define DERR(X) (derr << X) 
#define ASSERT(X) \
do { \
  if(!(X)) { \
    cerr << "Internal Error, In function " << derr.function_name() << endl; \
    cerr << "Assertion \"" << #X << "\" failed!" << endl; \
    cerr << "in file \"" << __FILE__ << "\"" << endl; \
    cerr << "at line " << __LINE__ << endl; \
    derr << "Assertion \"" << #X << "\" failed!" << endl; \
    exit(1); \
  } \
} while(0)

class DynaDebug;

typedef DynaDebug& (*DynaDebugFunc)(DynaDebug &);

DynaDebug& endl(DynaDebug& derr);

class DynaDebug 
{
public:

    enum { MAX_LABEL_LENGTH = 128 };
    DynaDebug(const char *, int);
    ~DynaDebug();
    static bool IsDebuggingOn(int);

    // char
    DynaDebug &operator<<(const char c);
    DynaDebug &operator<<(const signed char c);
    DynaDebug &operator<<(const unsigned char c);

    // char *
    DynaDebug &operator<<(const char * s);
    DynaDebug &operator<<(const signed char * s);
    DynaDebug &operator<<(const unsigned char * s);
    DynaDebug &operator<<(const XString & s);

    // short
    DynaDebug &operator<<(const signed short i);
    DynaDebug &operator<<(const unsigned short i);

    // int
    DynaDebug &operator<<(const signed int i);
    DynaDebug &operator<<(const unsigned int i);

    // long
    DynaDebug &operator<<(const signed long i);
    DynaDebug &operator<<(const unsigned long i);

    // flog, double, long double
    DynaDebug &operator<<(const float f);
    DynaDebug &operator<<(const double d);
    DynaDebug &operator<<(const long double d);

    // function
    DynaDebug &operator<<(DynaDebugFunc f);

    // function name
    const char * function_name(void);

    static char *config_filename;
    static char *output_filename;

private:

    friend DynaDebug& endl(DynaDebug& os);

    DynaDebug();                       // Not defined, cannot use
    DynaDebug(const DynaDebug &);      // Not defined, cannot use

    static void Initialize(void);
    static void FreeContext(void *);
    static void UpdateCache(void);
    static void LoadCache(void);

    void indent(void);

    static int num_labels;
    static char **labels;
    static bool *cache; 
    static int indent_size;
    static char *environment_flagname;
    static bool init_okay;

    static ofstream derr;

    struct DebugContext
    {
        public:

        enum { CBUF_SIZE = 16 };

        char cbuf[CBUF_SIZE];
        XString sbuf;

        int cur_indent; 
        bool indent_flag;

#ifdef _REENTRANT

        int thread_number;

#endif

    };

#ifdef _REENTRANT

    static pthread_once_t debug_init_once;    
    static pthread_key_t debug_data_key;
    static pthread_mutex_t debug_mutex;
    static int thread_count;
 
#else

    static bool first_time;
    static DebugContext single_context;

#endif

    DebugContext *context;
    const char *curname;
    bool debug_flag;

};

#endif
#endif
