/* File:  InetConnection.cpp 

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is the InetConnection class/library

The Initial Developer of the Original Code is Gregory Riggs.
Portions created by Gregory Riggs are Copyright (C) 2000, 2001
Gregory Riggs.  All Rights Reserved.

Contributor(s):
  Gregory Riggs <griggs@itqlata.com> <greg@nmia.com>
  Jerry Karasz <jkarasz@itqlata.com>
  ITQ LATA, LLC. <wctpxml@itqlata.com>

*/

#include "InetConnection.hpp"
#include "DynaDebug.hpp"
#include "DebugLabels.hpp"

bool InetConnection::abort_flag = false;

//----------------------------------------------------------------------------
// InetConnection
//----------------------------------------------------------------------------

InetConnection::InetConnection()
{
    DOBJ("InetConnection::InetConnection()", INET_CONN_DEBUG);

    sock_fd = -1;
    const long int usec = 10000000L; 
    polling_interval = usec/2;
    session_limit = 5;  // 5 seconds
    cancel_flag = false;
}

//----------------------------------------------------------------------------
// ~InetConnection
//----------------------------------------------------------------------------

InetConnection::~InetConnection()
{
    DOBJ("InetConnection::~InetConnection()", INET_CONN_DEBUG);

    fast_close();
}

//----------------------------------------------------------------------------
// is_closed
//----------------------------------------------------------------------------

bool
InetConnection::is_closed(void)
{
    DOBJ("InetConnection::is_closed()", INET_CONN_DEBUG);

    if(sock_fd < 0)
    {
        DERR("socket is closed" << endl);
        return true;
    }
    DERR("socket is open" << endl);
    return false;
}

//----------------------------------------------------------------------------
// InetConnection::make_nonblocking
//----------------------------------------------------------------------------

int
InetConnection::make_nonblocking(
    XString& error)
{
    DOBJ("InetConnection::make_nonblocking()", INET_CONN_DEBUG);

    //
    // Get Socket Flags
    //
    DERR("before fcntl" << endl);
    int flags = fcntl(sock_fd, F_GETFL, 0);
    if(flags < 0)
    {
        int errnum = errno;
        error = translate_error(errnum);
        DERR("fcntl error: " << error << endl);
        fast_close();
        return INET_CONN_ERROR;
    }

    //
    // Make socket non-blocking
    //
    DERR("before fcntl" << endl);
    int res = fcntl(sock_fd, F_SETFL, flags | O_NONBLOCK);
    if(res < 0)
    {
        int errnum = errno;
        error = translate_error(errnum);
        DERR("fcntl error: " << error << endl);
        fast_close();
        return INET_CONN_ERROR;
    }

    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// put_conn
//----------------------------------------------------------------------------

int
InetConnection::put_fd(
    XString& error, 
    int fd,
    const sockaddr_in& addr)
{
    client_addr = addr;
    sock_fd = fd;

    //
    // Make socket non-blocking
    //
    int error_code = make_nonblocking(error);
    if(error_code != INET_CONN_OKAY)
    {
        return error_code;
    }

    //
    // Record session start time
    //
    session_start = time(NULL);

    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// open_conn
//----------------------------------------------------------------------------

int
InetConnection::open_conn(
    XString& error,
    const XString& hostname,
    int port)
{
    DOBJ("InetConnection::open_conn()", INET_CONN_DEBUG);

    int error_code;
    struct sockaddr_in server_addr;

    //
    // Set IP Address
    //
    error_code = lookup_ipaddr(error, hostname, &server_addr);
    if(error_code != INET_CONN_OKAY)
    {
        sock_fd = -1;
        return error_code;
    }

    //
    // Set Port
    //
    server_addr.sin_family = AF_INET;
    server_addr.sin_port = htons(port);

    //
    // Create a socket
    //
    DERR("before socket" << endl);
    sock_fd = socket(AF_INET, SOCK_STREAM, 0);
    if(sock_fd < 0)
    {
        int errnum = errno;
        error = translate_error(errnum);
        DERR("socket error: " << error << endl);
        sock_fd = -1;
        return INET_CONN_ERROR;
    }

    //
    // Make socket non-blocking
    //
    error_code = make_nonblocking(error);
    if(error_code != INET_CONN_OKAY)
    {
        return error_code;
    }

    //
    // Recored session start
    //
    session_start = time(NULL);

    //
    // Connect socket to the server
    //
    DERR("before connect" << endl);
    int res = connect(sock_fd, 
        (struct sockaddr *)&server_addr, sizeof(server_addr));
    if(res >= 0)
    {
        // connected immediately
        return INET_CONN_OKAY;
    }

    int errnum = errno; 
    if(errnum != EINPROGRESS)
    {
        error = translate_error(errnum);
        DERR("connect error: " << error << endl);
        return INET_CONN_ERROR;
    }

    //
    // Wait for connection to server to be established
    //
    error_code = ready_for_io(error,true,true);
    if(error_code != INET_CONN_OKAY)
    {
        return error_code;
    }

    //
    // Check if connection was established okay
    //
    DERR("before getsockopt" << endl);
    errnum = 0;
    socklen_t errlen = sizeof(errnum);
    res = getsockopt(sock_fd, SOL_SOCKET, SO_ERROR, &errnum, &errlen);
    if(res < 0) 
    {
        errnum = errno;
        error = translate_error(errnum);
        DERR("getsockopt error: " << error << endl);
        fast_close();
        return INET_CONN_ERROR;
    }
    else if(errnum != 0)
    {
        error = translate_error(errnum);        
        fast_close();
        DERR("connect error: " << error << endl);
        return INET_CONN_ERROR;
    }

    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// close_conn
//----------------------------------------------------------------------------

int
InetConnection::close_conn(XString& error)
{
    DOBJ("InetConnection::close_conn()", INET_CONN_DEBUG);

    //
    // Shutdown the socket connection
    //

    DERR("before shutdown" << endl);
    int res = shutdown(sock_fd, 2);
    if(res < 0)
    {
        int errnum = errno;
        error = translate_error(errnum);        
        DERR("shutdown error: " << error << endl);
        fast_close();
        return INET_CONN_ERROR;
    }

    //
    // Wait for the buffers to flush
    //

    int error_code = ready_for_read(error);
    if(error_code != INET_CONN_OKAY)
    {
        return error_code;
    }

    //
    // Now close the socket
    //

    DERR("before close" << endl);
    res = close(sock_fd);
    sock_fd = -1;
    if(res < 0)
    {
        int errnum = errno;
        error = translate_error(errnum);        
        DERR("close error: " << error << endl);
        return INET_CONN_ERROR;
    }
    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// fast_close
//----------------------------------------------------------------------------

void
InetConnection::fast_close(void)
{
    DOBJ("InetConnection::fast_close()", INET_CONN_DEBUG);

    if(sock_fd < 0)
    {
        return;
    }

    DERR("closing sock_fd=" << sock_fd << endl);

    XString error;
    int res = close(sock_fd);
    sock_fd = -1;
    if(res < 0)
    {
        int errnum = errno;
        error = translate_error(errnum);        
        DERR("close error: " << error << endl);
    }
    return;
}

//----------------------------------------------------------------------------
// ready_for_io
//----------------------------------------------------------------------------

int
InetConnection::ready_for_io(
    XString& error, 
    bool read_flag, 
    bool write_flag)
{
    DOBJ("InetConnection::ready_for_io()", INET_CONN_POLL_DEBUG);

    int count = 0;
    bool done = false;
    while(!done)
    {
        if(cancel_flag || abort_flag)
        {
            DERR("cancel/abort flag set" << endl);
            error = "cancel/abort flag set";
            fast_close();
            return INET_CONN_CANCEL;
        }

        if(session_limit > 0) 
        {
            time_t current_time = time(NULL);
            time_t session_time = current_time - session_start;
            DERR("session_time=" << session_time << ", " <<
                 "session_limit=" << session_limit << endl);
            if(session_time > session_limit)
            {
                error = "session timeout";
                DERR("session timeout" << endl);
                fast_close();
                return INET_CONN_TIMEOUT;
            }
        }

        const long int usec = 10000000L;
        struct timeval delay;
        delay.tv_sec = polling_interval / usec;
        delay.tv_usec = polling_interval % usec;
        DERR("deltime sec=" << delay.tv_sec << ", "
             "usec=" << delay.tv_usec << endl);

        fd_set read_fds, write_fds;
        fd_set *p_read_fds, *p_write_fds;
     
        p_read_fds = NULL;
        if(read_flag)
        { 
            DERR("select wait for read" << endl);
            p_read_fds = &read_fds;
            FD_ZERO(p_read_fds);
            FD_SET(sock_fd, p_read_fds);
        }
 
        p_write_fds = NULL;
        if(write_flag)
        {
            DERR("select wait for write" << endl);
            p_write_fds = &write_fds;
            FD_ZERO(p_write_fds);
            FD_SET(sock_fd, p_write_fds);
        }

        DERR("before select, count=" << count++ << endl);
        int res = select(sock_fd+1, p_read_fds, p_write_fds, NULL, &delay);
        if(res < 0) 
        {
            int errnum = errno;
            if(errnum != EINTR)
            {
                error = translate_error(errnum);
                DERR("select error: " << error << endl);
                fast_close();
                return INET_CONN_ERROR;
            }
        }
        else if(res > 0)
        {
            done = true;
        }
    }
    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// ready_for_read
//----------------------------------------------------------------------------

int
InetConnection::ready_for_read(
    XString& error)
{
    DOBJ("InetConnection::ready_for_read()", INET_CONN_DEBUG);

    return ready_for_io(error, true, false);
}

//----------------------------------------------------------------------------
// ready_for_write
//----------------------------------------------------------------------------

int
InetConnection::ready_for_write(
    XString& error)
{
    DOBJ("InetConnection::ready_for_write()", INET_CONN_DEBUG);

    return ready_for_io(error, false, true);
}

//----------------------------------------------------------------------------
// writen
//----------------------------------------------------------------------------

int
InetConnection::writen(
    XString &error,
    const char *data,
    int length)
{
    DOBJ("InetConnection::writen()", INET_CONN_DEBUG);

    int count = 0;
    while(count < length)
    {
        DERR("count=" << count << ", length=" << length << endl);

        int error_code = ready_for_write(error);
        if(error_code != INET_CONN_OKAY)
        {
            return error_code;
        }

        int write_len = length - count;
        DERR("before write, write_len=" << write_len << endl);
        int res = write(sock_fd, data + count, write_len);
        if(res < 0)
        {
            int errnum = errno;
            if(errnum == EPIPE)
            {
                DERR("got EPIPE" << endl);
                error="client closed connection";
                fast_close();
                return INET_CONN_CLOSE;
            }
            else if(errnum != EINTR && errnum != EAGAIN)
            {
                error = translate_error(errnum);
                DERR("write error: " << error << endl);
                fast_close();
                return INET_CONN_ERROR;
            }
        }

        count += res;
    }

    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// readn
//----------------------------------------------------------------------------

int 
InetConnection::readn(
    XString &error,
    XString &data,
    int length)
{
    DOBJ("InetConnection::readn()",INET_CONN_DEBUG);

    data = "";
    int count = 0;
    while(count < length) 
    {
        DERR("count=" << count << ", length=" << length << endl);

        int error_code = ready_for_read(error);
        if(error_code != INET_CONN_OKAY)
        {
            return error_code;
        }

        int read_len = length - count;
        if(read_len > BUFFER_SIZE)
        {
            read_len = BUFFER_SIZE;
        }

        DERR("before read, read_len=" << read_len << endl);
        int res = read(sock_fd, buffer, read_len);
        if(res < 0)
        {
            int errnum = errno;
            if(errnum != EINTR && errnum != EAGAIN)
            {
                error = translate_error(errnum);
                DERR("read error: " << error << endl);
                fast_close();
                return INET_CONN_ERROR;
            }
        }
        else if(res == 0)
        {
            DERR("read returned 0, connection closed" << endl);
            error = "client closed connection";
            fast_close();
            return INET_CONN_CLOSE;
        }
        else
        {
            data.append(buffer, res);
            count += res;
        }
    }

    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// read_available
//----------------------------------------------------------------------------
    
int
InetConnection::read_available(XString& error, XString& data)
{
    DOBJ("InetConnection::read_available()", INET_CONN_DEBUG);

    int error_code = ready_for_read(error);
    if(error_code != INET_CONN_OKAY)
    {
        return error_code;
    }

    DERR("before read" << endl);
    int res = read(sock_fd, buffer, BUFFER_SIZE);
    if(res > 0)
    {
        XString xtmp(buffer,res);
        DERR("read data |" << xtmp.c_str() << "|" << endl);
        data.append(buffer, res);
    }
    else if(res == 0)
    {
        DERR("read returned 0, connection closed" << endl);
        error = "client closed connection";
        fast_close();
        return INET_CONN_CLOSE;
    }
    else if(res < 0)
    {
        int errnum = errno;
        if(errnum != EINTR && errnum != EAGAIN)
        {
            error = translate_error(errnum);
            DERR("read error: " << error << endl);
            fast_close();
            return INET_CONN_ERROR;
        }
    }

    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// lookup_ipaddr
//----------------------------------------------------------------------------

int
InetConnection::lookup_ipaddr(
    XString& error,
    const XString& hostname,
    struct sockaddr_in *addr)
{
    DOBJ("InetConnection::lookup_ipaddr()", INET_CONN_DEBUG);

    struct hostent *hentp;
    int herrno;
    int res;

    //
    // First see if we are already given the IP address
    // as an ascii dotted-decimal string
    //
    res = inet_aton(hostname.c_str(), &(addr->sin_addr));  
    if(res)
    {
        return INET_CONN_OKAY;
    }

    //
    // Perform hostname to IP address lookup
    // 

#ifdef _REENTRANT

    struct hostent hent;

    // Warning Linux specific code here, 

    const int hbuf_size = 8192;
    char hbuf[hbuf_size];
    
    res = gethostbyname_r(hostname.c_str(), &hent, 
        hbuf, hbuf_size, &hentp, &herrno);

    if(res < 0)
    {
        hentp = NULL;
    }

#else

    hentp = gethostbyname(hostname.c_str());
    herrno = h_errno;

#endif

    if(hentp == NULL)
    {
        error = translate_host_error(herrno);
        return INET_CONN_ERROR;
    }

    memset(addr, 0, sizeof(struct sockaddr_in));
    memcpy(&(addr->sin_addr), hentp->h_addr, hentp->h_length);

    return INET_CONN_OKAY;
}

//----------------------------------------------------------------------------
// translate_error
//----------------------------------------------------------------------------

XString 
InetConnection::translate_error(int errnum)
{
    DOBJ("InetConnection::translate_error()", INET_CONN_DEBUG);

    char buf[16];
    XString error;
    sprintf(buf,"%d",errnum);

    // GKR TODO add error string translations
    // for now just put the errno number

    error = "errno = ";
    error += buf;

    // JDK Added 3/13/01
    error += ", see errno.h for a definition.\n";

    return error;
}

//----------------------------------------------------------------------------
// translate_host_error
//----------------------------------------------------------------------------

XString 
InetConnection::translate_host_error(int herrnum)
{
    DOBJ("InetConnection::translate_host_error()", INET_CONN_DEBUG);

    char buf[16];
    XString error;
    sprintf(buf,"%d",herrnum);

    // GKR TODO add error string translations
    // for now just put the h_errno number

    error = "h_errno=";
    error += buf;
    return error;
}
