#!/usr/bin/perl
# 
# File:  UpdateDebug.pl
#
# The contents of this file are subject to the Mozilla Public
# License Version 1.1 (the "License"); you may not use this file
# except in compliance with the License. You may obtain a copy of
# the License at http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS
# IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
# implied. See the License for the specific language governing
# rights and limitations under the License.
#
# The Original Code is DynaDebug class/library
#
# The Initial Developer of the Original Code is Gregory Riggs.
# Portions created by Gregory Riggs are Copyright (C) 1996, 2000, 2001
# Gregory Riggs.  All Rights Reserved.
#
# Contributor(s):
#   Gregory Riggs <griggs@itqlata.com> <greg@nmia.com>
#   ITQ LATA, LLC. <wctpxml@itqlata.com>
#

#
# This perl script will go through all of the "*.cpp" files in 
# the current directory and search them for DYNADEBUG labels.  
# It will then update the following three files 
# "DebugLabels.cpp", "DebugLabels.hpp", and "debug.cfg" 
# so that they will reflect what DYNADEBUG labels are being used.  
# This automates a very tedious chore of finding and sorting
# the labels, and putting them in their needed format for those files.
#
 
$debugbase = "DebugLabels";

$hfile = "$debugbase.hpp";
$cfile = "$debugbase.cpp";
$dfile = "debug.cfg";
$ofile = "debug.out";
$eflag = "DYNADEBUG";
$isize = 4;

$MAX_LABEL_LENGTH = 32;

&get_labels();

$comment = "//";
&update_file($cfile);
&update_file($hfile);
$comment = "#";
&update_file($dfile);

#---------------------------------------------------------------------------
# sub update_htype 
#---------------------------------------------------------------------------

sub update_htype {
  
  @new_middle = ();

  push(@new_middle,"#include \"DynaDebug.hpp\"\n");
  push(@new_middle,"#ifndef DEBUGLABELS_H\n");
  push(@new_middle,"#define DEBUGLABELS_H\n");
  push(@new_middle,"\n");

  $count = 0;
  foreach $label (@labels) {
    $line  = "#define $label ";
    $line .= " " x ($MAX_LABEL_LENGTH - length($label));
    $line .= "$count\n";
    push(@new_middle,$line);
    $count ++; 
  }

  $label = "NUMBER_DEBUG_LABELS";
  $line  = "#define $label ";
  $line .= " " x ($MAX_LABEL_LENGTH - length($label));
  $line .= "$count\n";
  push(@new_middle,$line);
  push(@new_middle,"\n");
  push(@new_middle,"#endif\n");

}

#---------------------------------------------------------------------------
# sub update_ctype 
#---------------------------------------------------------------------------

sub update_ctype {

  local($line,$label);

  foreach $line (@old_middle) {
    if($line =~ /indent_size\s*\=\s*(\d+)/) {
      $isize = $1;
    }
    if($line =~ /environment_flagname\s*\=\s*\"(.*)\"/) {
      $eflag = $1;
    }
    if($line =~ /config_filename\s*\=\s*\"(.*)\"/) {
      $dfile = $1;
    }
    if($line =~ /output_filename\s*\=\s*\"(.*)\"/) {
      $ofile = $1;
    }
  }

  @new_middle = ();

  push(@new_middle,"#ifdef DYNADEBUG\n");
  push(@new_middle,"#include \"DebugLabels.hpp\"\n");
  push(@new_middle,"\n");
  push(@new_middle,"char *DebugLabels[] =\n");
  push(@new_middle,"{\n");

  foreach $label (@labels) {
    $line  = "  \"$label\",\n";
    push(@new_middle,$line);
  }
  push(@new_middle,"  NULL\n");

  push(@new_middle,"};\n");
  push(@new_middle,"\n");
  push(@new_middle,"bool DebugCache[NUMBER_DEBUG_LABELS];\n");
  push(@new_middle,"\n");
  push(@new_middle,"int DynaDebug::num_labels = NUMBER_DEBUG_LABELS;\n");
  push(@new_middle,"char **DynaDebug::labels = DebugLabels;\n");
  push(@new_middle,"bool *DynaDebug::cache = DebugCache;\n");
  push(@new_middle,"int DynaDebug::indent_size = $isize;\n");
  push(@new_middle,"char *DynaDebug::environment_flagname = \"$eflag\";\n");
  push(@new_middle,"char *DynaDebug::config_filename = \"$dfile\";\n");
  push(@new_middle,"char *DynaDebug::output_filename = \"$ofile\";\n");
  push(@new_middle,"\n");
  push(@new_middle,"#endif\n");
}

#---------------------------------------------------------------------------
# sub update_dtype 
#---------------------------------------------------------------------------

sub update_dtype {

  @new_middle = ();
  push(@new_middle,"\n");

  foreach $label (@labels) {

    $found = 0;

    foreach $line (@old_middle) {
      if($line =~ /^$label\s*$/){
        $found = 1; 
      } elsif($line =~ /^\#\s*$label\s*$/) {
        $found = 1;
      }
      if($found) {
        push(@new_middle,$line);
        last;
      }
    } 

    if(!$found){
      push(@new_middle,"$label\n");
    }

  }

  push(@new_middle,"\n");

}

#---------------------------------------------------------------------------
# sub cmp_middles 
#---------------------------------------------------------------------------
# returns true if the middles are equal;

sub cmp_middles {

  $oldnum = @old_middle;
  $newnum = @new_middle;
  $res = 1;
  if($oldnum == $newnum) {
    for($i = 0; $i < $oldnum; $i++) {
      if($old_middle[$i] ne $new_middle[$i]){
        $res = 0;
        last;        
      } 
    }
  } else {
    $res = 0;
  }
  return $res;

}

#---------------------------------------------------------------------------
# sub update_file 
#---------------------------------------------------------------------------

sub update_file {

  local($file) = @_;

  &pre_scan($file);

  if($file =~ /\.hpp$/) {
    &update_htype();
  } elsif ($file =~ /\.cpp$/) {
    &update_ctype();
  } else {
    &update_dtype();
  }

  if(&cmp_middles()) {
   
    print "File \"$file\" is already up to date.\n";

  } else {

    print "Updating the file \"$file\"\n";
    open(FILE,">$file") ||
    die("Couldn't open \"$hfile\" for write");

    foreach $line (@before,@new_middle,@after) {
      print FILE $line;
    }
    close(FILE);

  }

}

#---------------------------------------------------------------------------
# sub get_labels 
#---------------------------------------------------------------------------

sub get_labels {

  local($name);
  local(@files);

  opendir(CURDIR,".") || die ("No current directory?");
  while($name = readdir(CURDIR)) {
    if(($name =~ /\.cpp$/) && 
       ($name ne $cfile) && 
       ($name ne "DynaDebug.cpp")
      ){
      push(@files, $name);
    }
  }

  foreach $file (@files) {
    open(FILE,"$file") ||
      die("Couldn't open \"$file\" for read");
    while(<FILE>){
      if(/DOBJ\(.*\,\s*(\w+\_DEBUG)\s*\)/) {
        $labels{$1} = "off";
      }
    }  
    close(FILE);
  }

  @labels = sort (keys %labels); 
}

#---------------------------------------------------------------------------
# sub pre_scan 
#---------------------------------------------------------------------------

sub pre_scan {

  local($file) = @_;
  @before = ();
  @old_middle = ();
  @after = ();

  if(open(FILE,"$file")) {

    $before = 0;
    $after = 0;

    while(<FILE>)
    {
      if(!$before) {
        push(@before,$_);
        if(/Begin DynaDebug Automation Region/) {
          $before = 1;
        } 
      } elsif(!$after) {
        if(/End DynaDebug Automation Region/) {
          push(@after,$_);
          $after = 1;
        } else {
          push(@old_middle,$_);
        }
      } else {
        push(@after,$_);
      }
    }

    if(!$before) {
      die("No \"Begin DynaDebug Automation Region\" in \"$file\""); 
    } elsif(!$after) {
      die("No \"End DynaDebug Automation Region\" in \"$file\""); 
    }

    close(FILE);

  } else {

    if(open(FILE,">$file")) {

      print "Creating \"$file\"...\n";

      @before = "$comment Begin DynaDebug Automation Region\n";
      @after = "$comment End DynaDebug Automation Region\n";
      print FILE $before;
      print FILE $after;
      close(FILE);

    } else {

      die("Error, couldn't open \"$file\" for write");

    }
  }
}

