/* File:  wctpConnection.cpp

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.skytel.com/downloads/wctpxml/worldcom_mpl_1.1.txt.

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is the WCTP Toolkit

The Initial Developer of the Original Code is WorldCom, Inc.
Portions created by WorldCom, Inc. are Copyright  2001
WorldCom, Inc.  All Rights Reserved.

Contributor(s):
  WorldCom, Inc. <skytel-strategic-apps@wcom.com>  

*/

#include "wctpConnection.hpp"
#include "EncodeUtil.hpp"

//----------------------------------------------------------------------------
// wctpConnection
//----------------------------------------------------------------------------

wctpConnection::wctpConnection()
{
    m_leftover_raw = "";
    m_leftover_xml = "";
    m_send_urlencode = WCTP_URL_ENCODE_NONE;
    m_receive_urlencode = WCTP_URL_ENCODE_NONE;
}

//----------------------------------------------------------------------------
// ~wctpConnection
//----------------------------------------------------------------------------

wctpConnection::~wctpConnection()
{
}

//----------------------------------------------------------------------------
// set_send_urlencode
//----------------------------------------------------------------------------

void 
wctpConnection::set_send_urlencode(wctp_url_encode_t send_urlencode)
{
    m_send_urlencode = send_urlencode;
}

//----------------------------------------------------------------------------
// get_send_urlencode
//----------------------------------------------------------------------------

wctp_url_encode_t 
wctpConnection::get_send_urlencode(void)
{
    return m_send_urlencode;
}

//----------------------------------------------------------------------------
// set_receive_urlencode
//----------------------------------------------------------------------------

void 
wctpConnection::set_receive_urlencode(wctp_url_encode_t receive_urlencode)
{
    m_receive_urlencode = receive_urlencode;
}

//----------------------------------------------------------------------------
// get_receive_urlencode
//----------------------------------------------------------------------------

wctp_url_encode_t
wctpConnection::get_receive_urlencode(void)
{
    return m_receive_urlencode;
}

//----------------------------------------------------------------------------
// wctp_encode
//----------------------------------------------------------------------------

void
wctpConnection::wctp_encode(XString& dest, const XString& src)
{
    if(m_send_urlencode == WCTP_URL_ENCODE_FULL)
    {
        EncodeUtil::url_encode(dest, src);
    }  
    else if(m_send_urlencode == WCTP_URL_ENCODE_PART)
    {
        EncodeUtil::partial_url_encode(dest, src);
    }
    else
    {
        dest = src;
    }
}

//----------------------------------------------------------------------------
// wctp_decode
//----------------------------------------------------------------------------

void
wctpConnection::wctp_decode(XString& dest, XString& src)
{
    if(m_receive_urlencode == WCTP_URL_ENCODE_FULL)
    { 
        EncodeUtil::url_decode(dest, src);
    }
    else if(m_receive_urlencode == WCTP_URL_ENCODE_PART)
    {
        EncodeUtil::partial_url_decode(dest, src);
    }
    else
    {
        dest += src;
        src = "";
    }
}

//----------------------------------------------------------------------------
// write_xml
//----------------------------------------------------------------------------

int 
wctpConnection::write_xml(XString &error, const XString &xml)
{
    XString out;
    wctp_encode(out,xml);
    return writen(error, out.data(), out.length());
}

//----------------------------------------------------------------------------
// read_xml
//----------------------------------------------------------------------------

int 
wctpConnection::read_xml(XString &error, XString &xml)
{
    int pos, res;
    XString tag;
    bool done;

    pos = 0;
    xml = "";
    done = false;

    while(!done)
    {
        wctp_decode(m_leftover_xml, m_leftover_raw);        
        pos = complete_xml(xml, m_leftover_xml, tag, pos);
        if(!xml.empty())
        {
            res = INET_CONN_OKAY;
            return res; 
        }

        XString raw;
        res = read_available(error, raw); 
        m_leftover_raw += raw;
        if(res == INET_CONN_CLOSE)
        {
            wctp_decode(m_leftover_xml, m_leftover_raw);
            xml = m_leftover_xml;
            m_leftover_raw = "";
            m_leftover_xml = "";
            return res;
        }
        else if(res != INET_CONN_OKAY)
        {
            return res;
        }
    }

    return res;
}

//----------------------------------------------------------------------------
// is_tag_start_char
//----------------------------------------------------------------------------

bool
wctpConnection::is_tag_start_char(char ch)
{
    if(isalpha(ch) || ch == '_' || ch == ':')
    {
        return true;
    }
    return false;
}

//----------------------------------------------------------------------------
// is_tag_char
//----------------------------------------------------------------------------

bool
wctpConnection::is_tag_char(char ch)
{
    if(isalnum(ch) || ch == '_' || ch == ':' || ch == '-' || ch == '.')
    {
        return true;
    }
    return false;
}

//----------------------------------------------------------------------------
// locate_char_plus
//----------------------------------------------------------------------------

int
wctpConnection::locate_char_plus(
    const XString& src, 
    int start_pos, 
    char ch)
{
    int pos = start_pos;
    while(pos < src.size())
    {
        if(src[pos] == ch)
        {
            return pos+1;
        }
        pos++;
    }
    return start_pos; 
}

//----------------------------------------------------------------------------
// locate_tag
//----------------------------------------------------------------------------

int
wctpConnection::locate_tag(
    const XString& src, 
    int start_pos, 
    bool close_flag, 
    XString& tag)
{
    tag = "";  

    int stop_pos = locate_char_plus(src, start_pos,'>');
    if(stop_pos == start_pos)
    {
        return start_pos;
    }

    int pos = locate_char_plus(src,start_pos,'<');

    if(close_flag)
    {
        if(pos < src.size() && src[pos] == '/')
        {
            pos++;
        }
        else
        {
            return stop_pos;
        }
    }

    if(pos < src.size() && !is_tag_start_char(src[pos])) 
    {
        return stop_pos;
    }

    while(pos < src.size() && is_tag_char(src[pos]))
    {
        tag += src[pos];
        pos++;
    }

    return stop_pos;
}

//----------------------------------------------------------------------------
// complete_xml
//----------------------------------------------------------------------------

int
wctpConnection::complete_xml(
    XString &dest, 
    XString &src, 
    XString &tag, 
    int pos)
{
    dest = "";

    while(tag.empty())
    {
        int new_pos = locate_tag(src, pos, false, tag);
        if(new_pos == pos)
        {
            return pos;
        }
        pos = new_pos;
        if(!tag.empty())
        {
            // fprintf(stderr,"start tag is |%s|\n", tag.c_str());
        }
    }

    XString stop_tag;
    while(stop_tag != tag)   
    {
        int new_pos = locate_tag(src, pos, true, stop_tag);
        if(new_pos == pos)
        {
            return pos;
        }
        pos = new_pos;
        if(!stop_tag.empty())
        {
            // fprintf(stderr,"stop tag is |%s|\n", stop_tag.c_str());
        }
    }

    dest = src.left(pos);
    src = src.right(src.length()-pos);
    return true;
}
