/* File:  wctpXml_client.cpp

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.skytel.com/downloads/wctpxml/worldcom_mpl_1.1.txt.

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is the WCTP Toolkit

The Initial Developer of the Original Code is WorldCom, Inc.
Portions created by WorldCom, Inc. are Copyright  2001
WorldCom, Inc.  All Rights Reserved.

Contributor(s):
  WorldCom, Inc. <skytel-strategic-apps@wcom.com>  

*/

#include <stdlib.h>

#include "wctpConnection.hpp"
#include "DynaDebug.hpp"
#include "DebugLabels.hpp"

// WCTP files
#include "wctpXmlFailure.hpp"
#include "wctpXmlVersionQuery.hpp"
#include "wctpXmlVersionQueryResponse.hpp"
#include "wctpXmlVersionDTDSupport.hpp"
#include "wctpXmlVersionContactInfo.hpp"

#include "wctpXmlSubmitClientMessage.hpp"
#include "wctpXmlSubmitClientResponse.hpp"
#include "wctpXmlSubmitClientHeader.hpp"
#include "wctpXmlSubmitClientHeader.hpp"
#include "wctpXmlClientOriginator.hpp"
#include "wctpXmlClientMessageControl.hpp"
#include "wctpXmlRecipient.hpp"
#include "wctpXmlPayload.hpp"
#include "wctpXmlAlphanumeric.hpp"

// string used in searching for "start of response content"
// This is a hack for simplicity. Do it right, and get a
// real HTTP package
const XString xml_start = "<?xml";

#define BOOL  char
#define FALSE 0
#define TRUE  (!FALSE)

///////////////////////////////////////////////////////////////
//
// Send a request to a server and get a response back,
//  regardless of what kind of WCTP request it was
//
///////////////////////////////////////////////////////////////

int 
do_request(
    XString                *pError_string, 
    const char             *host,
    const char             *port,
    const char             *path,
    wctpXmlOperation       *pRequest, 
    wctpXmlOperation       *pResponse)
{
    DOBJ("wctpXml_client's do_request()", WCTP_XML_CLIENT_DEBUG);


    wctpConnection conn;
    int            error_code = -1;  // generic non-INET failure code
    XString        assembled_request,
                   http_request,
                   http_response;
    char           buff[48];


    memset(buff,0,sizeof(buff));

    // all ptr args should be non-NULL
    if (pResponse     == NULL || 
        pError_string == NULL ||
        pRequest      == NULL ||
        path          == NULL ||
        port          == NULL ||
        host          == NULL)
    { // can't do that!
        DERR("ERROR - NULL Pointer" << endl);
        return  error_code;
    }

    // get info from request
    assembled_request = pRequest->AssembleRequestString();

    // put it into a simplistic HTTP 1.0 header 
    // First, we need the length of the assembled string
    int len = assembled_request.length();
    sprintf(buff,"%d",len);

    // Use a real HTTP package for this. This simplistic example shows the
    // important details, but is not robust or flexible
    http_request  = "POST ";
    http_request += path;
    http_request += " HTTP/1.0\n";
    http_request += "User-Agent: WCTP Open Source Lib Tester\n";
    http_request += "Host: ";
    http_request += host;
    http_request += "\n";
    http_request += "Content-Type: text/xml\n";
    http_request += "Content-Length: ";
    http_request += buff;
    http_request += "\n\n"; // The "extra" \n is REQUIRED by HTTP to signal 
                            // the end of the header/start of content
    http_request += assembled_request;

    // echo the request to stdout
    fprintf(stdout,"\n============================\n");
    fprintf(stdout,"THE HTTP REQUEST POSTED:\n");
    fprintf(stdout,"============================\n");
    fprintf(stdout,"%s\n", (const char *)http_request);

    // now open the connection and pass it along
    error_code = conn.open_conn(*pError_string, host, atoi(port));
    if (error_code != INET_CONN_OKAY)
    {
        DERR("ERROR - open_conn() failed" << endl);
        return error_code;
    }

    error_code = conn.write_xml(*pError_string, http_request);
    if (error_code != INET_CONN_OKAY)
    {
        DERR("ERROR - write_xml() failed" << endl);
        return error_code;
    }

    error_code = conn.read_xml(*pError_string, http_response);
    if (error_code != INET_CONN_OKAY)
    {
        DERR("ERROR - read_xml() failed" << endl);
        return error_code;
    }

    error_code = conn.close_conn(*pError_string);
    if (error_code != INET_CONN_OKAY)
    {
        DERR("ERROR - close_conn() failed" << endl);
        return error_code;
    }
   
    fprintf(stdout,"============================\n");
    fprintf(stdout,"THE HTTP RESPONSE RECEIVED:\n");
    fprintf(stdout,"============================\n");
    fprintf(stdout,"%s\n============================\n",
                                         (const char *) http_response);

    // Use a real HTTP package for this. This simplistic example shows the
    // important details, but is not robust or flexible
    int start = http_response.find(xml_start,0);
 
    if (start < 0)
    {   // can't parse it, no XML starting point
        DERR("ERROR - No XML start found" << endl);
        error_code = INET_CONN_ERROR;
    }
    else
    {   // okay, try to parse starting at "start"
        XString response_str = 
                    http_response.right(http_response.length() - start);
        if (pResponse->Parse(response_str) != WCTP_PARSE_OKAY)
        {   // can't parse it
            DERR("ERROR - Parse() failed" << endl);
            error_code = INET_CONN_ERROR;
        }
    }

    return error_code;
}

///////////////////////////////////////////////////////////////
//
// Query a server using wctp-VersionQuery
//
///////////////////////////////////////////////////////////////

void
do_VersionQuery(const char *host,
                const char *port,
                const char *path)
{
    DOBJ("wctpXml_client's do_VersionQuery()", WCTP_XML_CLIENT_DEBUG);


    wctpXmlVersionQuery         request;
    wctpXmlVersionQueryResponse response;
    XString                     error_string;
    int                         error_code = INET_CONN_ERROR;
    char                        buffer[256];


    // fill up the base class attributes
    request.SetVersion("wctp-dtd-v1r1");   // set to WCTP 1.1 

    // fill up the version query attributes
    request.SetInquirer("wctpXmlTester");

    // The GenerateTimestamp() call is NOT thread safe, but 
    // generates a time stamp in WCTP format from the current 
    // UTC time/GMT time
    const char *pTimeString = wctpXml::GenerateTimestamp();
    request.SetDateTime(pTimeString);

    // This call sends the request and gives us back the response
    // (we must delete it when we are done with it).
    error_code = do_request(&error_string, host, port, path, 
                                            &request, &response);
    if (error_code != INET_CONN_OKAY)
    {   // failed
        sprintf(buffer,"ERROR - error_code is %d", error_code);
        DERR(buffer << endl);
        sprintf(buffer,"ERROR - %s", error_string.c_str());
        DERR(buffer << endl);
        exit(0);
    }

    // we got something back as a response, check it out
    if (response.GetOperationType() != WCTP_XML_VERSION_QUERY_RESPONSE)
    {   // something wrong here!
        sprintf(buffer,"ERROR - Wrong Response Type %d", 
                     response.GetOperationType());
        DERR(buffer << endl);
        exit(0);
    }

    fprintf(stdout,"THE WCTP RESPONSE ATTRIBUTES:\n");
    fprintf(stdout,"============================\n");
    fprintf(stdout,"wctp-Operation Attributes:\n");
    fprintf(stdout,"\tVersion: %s\n", response.GetVersion());
    fprintf(stdout,"\tToken  : %s\n\n", response.GetToken());

    fprintf(stdout,"wctp-VersionResponse Attributes:\n");
    fprintf(stdout,"\tResponder         : %s\n", response.GetResponder());
    fprintf(stdout,"\tInquirer          : %s\n", response.GetInquirer());
    fprintf(stdout,"\tRequest datetime  : %s\n", response.GetDateTimeOfReq());
    fprintf(stdout,"\tResponse datetime : %s\n", response.GetDateTimeOfRsp());
    fprintf(stdout,"\tInvalid after     : %s\n", response.GetInvalidAfter());
    fprintf(stdout,"\tListDTDs          : %s\n", response.GetListDTD());
    fprintf(stdout,"\tListConfiguration : %s\n\n", response.GetListConfig());

    wctpXmlVersionContactInfo *pContactInfo = 
                        response.GetVersionContactInfo();
    if (pContactInfo != NULL)
    {   // have contact info, show it
        fprintf(stdout,"Contact Information:\n");
        fprintf(stdout,"\tEmail Address  : %s\n",pContactInfo->GetEmail());
        fprintf(stdout,"\tPhone Number   : %s\n",pContactInfo->GetPhone());
        fprintf(stdout,"\tWeb Address    : %s\n",pContactInfo->GetWWW());
        fprintf(stdout,"\tAdditional Info: %s\n\n",pContactInfo->GetInfo());
    }

    if (response.GetVersionQueryResponseType() == WCTP_XML_FAILURE)
    {   // Version Query failed
        fprintf(stdout,"VersionQuery FAILED\n");
        wctpXmlFailure *pFailure = response.GetFailure();
        if (pFailure != NULL)
        {
            fprintf(stdout,"Failure Information:\n");
            fprintf(stdout,"\tError Code     : %s\n",pFailure->GetErrorCode());
            fprintf(stdout,"\tError Text     : %s\n",pFailure->GetErrorText());
            fprintf(stdout,"\tAdditional Info: %s\n\n",pFailure->GetPcdata());
        }
    }
    else
    {   // Version Query succeeded
        fprintf(stdout,"VersionQuery SUCCEEDED\n");
        int num = response.NumVersionDTDSupport();
        for (int index = 0; index < num; index++)
        {   // get each DTDSupport object and show it
            wctpXmlVersionDTDSupport *pDTD = 
                                  response.GetVersionDTDSupport(index);
            if (pDTD != NULL)
            {   // show it
                fprintf(stdout,"DTD Information: %s\n",pDTD->GetDTDName());
                fprintf(stdout,"\tVersion Token: %s\n",pDTD->GetVerToken());
                fprintf(stdout,"\tSupport Type : %s\n",pDTD->GetSupportType());
                fprintf(stdout,"\tExceptions   : %s\n",pDTD->GetExceptions());
                fprintf(stdout,"\tSupport Until: %s\n",pDTD->GetSupportUntil());
                fprintf(stdout,"\tReplacement  : %s\n\n",
                                                        pDTD->GetReplacement());
            }
            else
            {
                fprintf(stdout,"No DTD Information available for DTD #%d\n",
                                 index);
            }
        }
    }
}

///////////////////////////////////////////////////////////////
//
// Send a Message using wctp-SubmitClientMessage
//
///////////////////////////////////////////////////////////////

void
do_SubmitClientMessage(const char *host,
                       const char *port,
                       const char *path,
                       const char *recip)
{
    DOBJ("wctpXml_client's do_SubmitClientMessage()", WCTP_XML_CLIENT_DEBUG);


    wctpXmlSubmitClientMessage  request;
    wctpXmlSubmitClientResponse response;
    XString                     error_string;
    int                         error_code = INET_CONN_ERROR;
    char                        buffer[256];


    if (recip         == NULL || 
        path          == NULL ||
        port          == NULL ||
        host          == NULL)
    { // can't do that!
        DERR("ERROR - NULL Pointer" << endl);
        return;
    }

    // fill up the base class attribute
    request.SetVersion("wctp-dtd-v1r1");   // set to WCTP 1.1 

    wctpXmlSubmitClientHeader *pHeader  = request.GetSubmitClientHeader();
    wctpXmlPayload            *pPayload = request.GetPayload();

    if (pHeader  == NULL || 
        pPayload == NULL)
    { // can't do that!
        DERR("ERROR - Unable to get ptr to header or payload" << endl);
        return;
    }

    // First, fill in message "body"
    pPayload->HoldAlphanumeric();
    wctpXmlAlphanumeric *pAlpha = pPayload->GetAlphanumeric();
    if (pAlpha  == NULL)
    { // can't do that!
        DERR("ERROR - Unable to get ptr to alpha payload" << endl);
        return;
    }
    pAlpha->SetPcdata("Test msg");

    // now for the Header fields
    // The GenerateTimestamp() call is NOT thread safe, but 
    // generates a time stamp in WCTP format from the current 
    // UTC time/GMT time
    const char *pTimeString = wctpXml::GenerateTimestamp();
    pHeader->SetSubmitTimestamp(pTimeString);

    pHeader->UseClientMessageControl();
    wctpXmlClientMessageControl *pCtrl  = pHeader->GetClientMessageControl();
    wctpXmlClientOriginator     *pOrig  = pHeader->GetClientOriginator();
    wctpXmlRecipient            *pRecip = pHeader->GetRecipient();

    if (pCtrl  == NULL || 
        pOrig  == NULL ||
        pRecip == NULL)
    { // can't happen!
        DERR("ERROR - Unable to get ptr to Ctrl, Orig, or Recip" << endl);
        return;
    }

    pCtrl->SetAllowTruncation(WCTP_BOOL_FALSE);
    pCtrl->SetDeliveryPriority(WCTP_PRIORITY_HIGH);

    pOrig->SetSenderID("WCTP Sender Test");
    pOrig->SetMiscInfo("WCTP Misc Info Test");

    pRecip->SetRecipientID(recip);

    // This call sends the request and gives us back the response
    // (we must delete it when we are done with it).
    error_code = do_request(&error_string, host, port, path, 
                                            &request, &response);

    if (error_code != INET_CONN_OKAY)
    {   // failed
        sprintf(buffer,"ERROR - error_code is %d", error_code);
        DERR(buffer << endl);
        sprintf(buffer,"ERROR - %s", error_string.c_str());
        DERR(buffer << endl);
        exit(0);
    }

    // we got something back as a response, check it out
    if (response.GetOperationType() != WCTP_XML_SUBMIT_CLIENT_RESPONSE)
    {   // something wrong here!
        sprintf(buffer,"ERROR - Wrong Response Type %d", 
                     response.GetOperationType());
        DERR(buffer << endl);
        exit(0);
    }

    fprintf(stdout,"THE WCTP RESPONSE ATTRIBUTES:\n");
    fprintf(stdout,"============================\n");
    fprintf(stdout,"wctp-Operation Attributes:\n");
    fprintf(stdout,"\tVersion: %s\n", response.GetVersion());
    fprintf(stdout,"\tToken  : %s\n\n", response.GetToken());

    if (response.GetSubmitClientResponseType() == WCTP_XML_FAILURE)
    {   // SubmitClientMessage failed
        fprintf(stdout,"SubmitClientMessage FAILED\n");
        wctpXmlFailure *pFailure = response.GetFailure();
        if (pFailure != NULL)
        {
            fprintf(stdout,"Failure Information:\n");
            fprintf(stdout,"\tError Code     : %s\n",pFailure->GetErrorCode());
            fprintf(stdout,"\tError Text     : %s\n",pFailure->GetErrorText());
            fprintf(stdout,"\tAdditional Info: %s\n\n",pFailure->GetPcdata());
        }
    }
    else
    {   // SubmitClientMessage succeeded
        fprintf(stdout,"SubmitClientMessage SUCCEEDED\n");
        wctpXmlClientSuccess *pSuccess = response.GetClientSuccess();
        if (pSuccess != NULL)
        {
            fprintf(stdout,"Success Information:\n");
            fprintf(stdout,"\tSuccess Code   : %s\n",
                                                   pSuccess->GetSuccessCode());
            fprintf(stdout,"\tSuccess Text   : %s\n",
                                                   pSuccess->GetSuccessText());
            fprintf(stdout,"\tTracking Number: %s\n",
                                                pSuccess->GetTrackingNumber());
            fprintf(stdout,"\tAdditional Info: %s\n\n",pSuccess->GetPcdata());
        }
    }
}

///////////////////////////////////////////////////////////////
// usage
///////////////////////////////////////////////////////////////

void usage(void)
{
    fprintf(stderr,"\n\nUsage: wctpXml_client -h <host> -p <port>");
    fprintf(stderr," -a <path> [-r <recipient>]\n");
    fprintf(stderr,"For example:\n");
    fprintf(stderr,"    wctpXml_client -h 123.234.321.2 -p 80 -a /wctp\n\n");
    fprintf(stderr,"         or:\n\n");
    fprintf(stderr,"    wctpXml_client -h 123.234.321.2 -p 80 -a /wctp -r ");
    fprintf(stderr,"1234567\n\n");
    exit(0);
}

///////////////////////////////////////////////////////////////
// main
///////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
    DOBJ("wctpXml_client's main()", WCTP_XML_CLIENT_DEBUG);

    char *host     = NULL;
    char *path     = NULL;
    char *port     = "80";    // default, HTTP port
    char *recip    = "";      // default, no recip and no msg being sent
    int   i        = 0;

    BOOL  bGotHost = FALSE;
    BOOL  bGotPort = FALSE;
    BOOL  bGotPath = FALSE;
    BOOL  bGotRecip= FALSE;

    //
    // Get host and port number from command line arguments
    //
    if (argc < 7) 
    {   // not enough args, show usage and quit
        usage();
    }

    for (i = 1; i < argc; i++)
    {
        if (strcmp(argv[i],"-p")==0)
        {   // port on host
            i++;
            if (i < argc)
            {
                port = argv[i];                
                bGotPort = TRUE;
            }
            else
            {
                usage();
            }
        }
        else if (strcmp(argv[i],"-h")==0)
        {   // host address
            i++;
            if (i < argc)
            {
                host = argv[i];
                bGotHost = TRUE;
            }
            else
            {
                usage();
            }
        }
        else if (strcmp(argv[i],"-a")==0)
        {   // path
            i++;
            if (i < argc)
            {
                path = argv[i];
                bGotPath = TRUE;
            }
            else
            {
                usage();
            }
        }
        else if (strcmp(argv[i],"-r")==0)
        {   // recipient
            i++;
            if (i < argc)
            {
                recip = argv[i];
                bGotRecip = TRUE;
            }
            else
            {
                usage();
            }
        }
        else
        {
            usage();
        }
    }

    if (!bGotHost ||
        !bGotPort ||
        !bGotPath)
    {   // need them all to proceed!
        DERR("ERROR - Insufficient arguments!" << endl);
        usage(); // quit
    }

    fprintf(stdout,"Connecting using host  =|%s|\n",host);
    fprintf(stdout,"                 port  =|%s|\n",port);
    fprintf(stdout,"                 path  =|%s|\n",path);

    if (bGotRecip)
    {   // must send a message, not version query
        fprintf(stdout,"                 recip =|%s|\n\n",recip);
        do_SubmitClientMessage(host, port, path, recip);
    }
    else
    {   // must do version query, not send a message
        fprintf(stdout,"\n");
        do_VersionQuery(host, port, path);
    }

    DERR("DONE" << endl);
}
