/* File:  EncodeUtil.hpp

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.skytel.com/downloads/wctpxml/worldcom_mpl_1.1.txt.

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is the WCTP Toolkit

The Initial Developer of the Original Code is WorldCom, Inc.
Portions created by WorldCom, Inc. are Copyright  2001
WorldCom, Inc.  All Rights Reserved.

Contributor(s):
  WorldCom, Inc. <skytel-strategic-apps@wcom.com>  

*/

#ifndef ENCODE_UTIL_HPP
#define ENCODE_UTIL_HPP

//
// This utility class provides methods for encoding and decoding 
// using the BASE64 and UUENCODE formats.
//
// The BASE64 encoding/decoding format is defined in RFC1341 available at 
// http://sunsite.cnlab-switch.ch/ftp/doc/standard/rfc/13xx/1341
//
// A free public domain implementation of BASE64 written by
// John Walker is available at http://www.fourmilab.ch/webtools/base64/
// 
// The GNU uuencode/uudecode programs written by Ian Lance Taylor 
// are available in the GNU shareutils package available on GNU's ftp 
// server at ftp://ftp.gnu.org/gnu/sharutils/sharutils-4.2.1.tar.gz
// Note that the GNU GPL License does not allow including GPL 
// licensed code into a non-free program.
// 
// Note that the XString class can contain binary data, it allows
// null characters in the middle of the string.
//
// Note that uuencoded data is traditionally included between
// a "begin" and "end" line.  However, the examples in the WCTP
// specifications document (see examples on Page F-43 and F-45)
// indicate that the "begin" and "end" lines are not used. 
// This class currently does not add the "begin" and "end"
// lines or support parsing them.
//
//////////////////////////////////////////////////////////////
#include "XString.hpp"

class EncodeUtil
{
public:

    //////////////////////////////
    // base64 encode/decode
    static void
    base64_encode(XString &dest, const XString &src);
    static bool
    base64_decode(XString &dest, const XString &src);

    //////////////////////////////
    // uuencode/uudecode
    static void
    uuencode(XString &dest, const XString &src);
    static bool
    uudecode(XString &dest, const XString &src);

    //////////////////////////////
    // generates a hexdump 
    static void
    hexdump(XString &dest, const XString &src);

    //////////////////////////////
    // url encode/decode 
    static void
    url_decode(XString& dest, XString &src);
    static void
    url_encode(XString& dest, const XString& src);

    //////////////////////////////
    // partial url encode/decode
    static void
    partial_url_decode(XString& dest, XString &src);
    static void
    partial_url_encode(XString& dest, const XString& src);

    //////////////////////////////
    // Initialize the tables, 
    // (should be called once at startup if using multiple threads).
    static void 
    init_tables(void);

private:

    // in both operations below, pos is a zero-based index into "src"
    static bool uudecode_line(bool okay_flag, XString &dest, 
                              const XString &src, int pos);
    static char hex_to_char(const XString& src, int pos);

    static unsigned char base64_encode_table[64];
    static unsigned char base64_decode_table[256];
    static unsigned char uuencode_table[64];
    static unsigned char uudecode_table[256];

};

#endif
