/* File:  TStack.hpp

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is the TStack class/library

The Initial Developer of the Original Code is Gregory Riggs.
Portions created by Gregory Riggs are Copyright (C) 2001
Gregory Riggs.  All Rights Reserved.

Contributor(s):
  Gregory Riggs <greg@nmia.com>
  ITQ LATA, LLC. <wctpxml\@itqlata.com>  

*/

// This is a non-intrusive, indirect, expandable, non-memory owning,
// _T_raversable stack, implemented using a Double Linked list.  
// * It is non-intrusive in that it doesn't require the objects
//   stored on it to provide link fields.  
// * It is indirect in that it stores pointers to the objects instead
//   of storing a copy of the object. 
// * It is expandable in the sense that it doesn't have a maximum size
//   limit of how many items may be stored on the stack.  It allocates 
//   TStackNodes of the heap for the linkage fields.
// * It is non-memory owning in that it doesn't take ownership of the
//   items stored in it (with the exception of the special method destroy)
//
// Using various combinations of the shift, unshift, push, and pop 
// methods you can use it as either a stack or a queue.
//
// Using the operator[] function you can access the ith element in the
// stack.  It is this function that makes the stack traversable.  For a 
// single iterator, this operation is now O(1), but for nested iterators
// it is still O(n).
// 
// The push and unshift operators return a pointer to the TStackNode
// that was used to store the data element on the stack.  This pointer
// can then be used as a key to remove the data element from the double
// linked list in constant time.
//
// This TStack container class does not own the memory pointed at by
// the pointers stored in the stack.  However,  this class offers a
// special method "destroy" that will take ownership of the memory
// of all the objects stored in the stack and delete them.  This is offered
// for convenience so that the user of this stack doesn't have to delete
// each object on the stack seperately before letting the stack destruct.

#ifndef TSTACK_H
#define TSTACK_H

#include <stdlib.h>
#include <stddef.h>

class TStackBase;

//--------------------------------------------------------------------
// TStackNode class
//--------------------------------------------------------------------

class TStackNode 
{
public:
    void *data; 
private:
    TStackNode(void)
        : data(NULL), prev(NULL), next(NULL) {}
    friend class TStackBase;
    TStackNode *prev;
    TStackNode *next;
};

//--------------------------------------------------------------------
// TStackBase class
//--------------------------------------------------------------------

class TStackBase 
{
public:
    TStackBase();
    TStackBase(const TStackBase&);
    ~TStackBase();
    TStackBase& operator=(const TStackBase& ts);
    int size(void) const {return num;}
    TStackNode *push(void *d) {return add(sentry.prev,d);}
    TStackNode *unshift(void *d) {return add(&sentry,d);} 
    void *pop(void) {return remove(sentry.prev);}
    void *top(void) {return sentry.prev->data;}
    void *shift(void) {return remove(sentry.next);}
    TStackNode *operator[](int i);
    const TStackNode *operator[](int i) const;
    void *remove(TStackNode *);  
    void *remove(int *);
    void destroy(void (*del_func)(void *));

private:
    void Initialize(void);
    void Copy(const TStackBase& ts);
    TStackNode *add(TStackNode *p, void *d); 
    TStackNode sentry;
    int num;   // number of things on the stack
    int curpos;    // Keep track of last operator[] number 
    TStackNode *curptr;  // and last operator[] pointer
};

//--------------------------------------------------------------------
// template<class DATA> TStack class
//--------------------------------------------------------------------

template<class DATA> 
class TStack 
{
public: 

    // Returns the number of elements in the stack (same as length)
    int size(void) const
        { return stack.size(); }  

    // Returns the number of elements in the stack (same as size)
    int length(void) 
        { return stack.size(); }  

    // Returns true if the stack is empty
    bool empty(void) 
        { return stack.size() == 0; }

    // Adds to the top of the stack
    TStackNode *push(DATA *d)    
        { return stack.push((void *)d); }

    // Removes from the top of the stack
    DATA *pop(void)              
        { return (DATA *) stack.pop(); }

    // Returns the item on the top of the stack
    DATA *top(void) 
        { return (DATA *) stack.top(); }

    // Adds to the bottom of the stack
    TStackNode *unshift(DATA *d)     
        { return stack.unshift((void *)d); }

    // Removes from bottom of the stack
    DATA *shift(void)       
        { return (DATA *) stack.shift(); }

    // Returns the i_th item of the stack rvalue only,
    // if out of range returns NULL 
    DATA *operator[](int i) 
        { return (DATA *) stack[i]->data; } 
    const DATA *operator[](int i) const
        { return (DATA *) stack[i]->data; } 

    // Removes the i_th item from the stack
    DATA *remove(int i)    
        { return (DATA *) stack.remove(stack[i]); }

    // Removes the data pointed to by stack node
    DATA *remove(TStackNode *p)    
        { return (DATA *) stack.remove(p); }
 
    // Removes all items from the stack, but does not delete the data
    void clear(void)
        { stack.destroy(NULL); }

    // Takes ownership of all the data in the stack and deletes them
    void destroy(void) 
        { stack.destroy(dfunc); }

private:

    TStackBase stack;
    static void dfunc(void *d)
        { delete (DATA *)d; }
};

#endif
