/* File:  XString.hpp

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is the TStack class/library

The Initial Developer of the Original Code is Gregory Riggs.
Portions created by Gregory Riggs are Copyright (C) 2001
Gregory Riggs.  All Rights Reserved.

Contributor(s):
  Gregory Riggs <greg@nmia.com>
  ITQ LATA, LLC. <wctpxml\@itqlata.com>  

*/

//-----------------------------------------------------------------------------
//
// CHANGE HISTORY
// K. Houck - 5/8/2000 - Added has_white_space functions
//
// J. Poerschke - 11/7/00 - added a const char *find, Find, Left, and 
//    GetLength functions.  MFC's CString can now be typedef'd as 
//    XString if necessary.
//
//-----------------------------------------------------------------------------

#ifndef XSTRING_H
#define XSTRING_H

#define XSTRING_MAX_FORMAT_OUTPUT_STRING_LENGTH 500

class XString 
{
public:

    // Various constructors and copy constructors
    XString();                                 // XString x;
    XString(int init_memlen);                  // XString x(1024);
    XString(const XString &);                  // XString x(y);
    XString(const char *);                     // XString x("abc");
    XString(const char *, int size);           // XSTring x("abc",3);   
    XString(const char);                       // XString x('a');

    // destructor
    ~XString();

    // Assignment operators
    XString& operator=(const XString &);       // x = y;
    XString& operator=(const char *);          // x = "abc";
    XString& operator=(const char);            // x = 'a';

    // Concatenation operators
    XString& append(const char *, int);        // x.append("x",1);
    XString& operator+=(const XString &);      // x += y;
    XString& operator+=(const char *);         // x += "abc";
    XString& operator+=(const char);           // x += 'a';

    // string comparision operators
    friend int operator==(const XString&, const XString&);  // if(x == y)
    friend int operator!=(const XString&, const XString&);  // if(x != y)
    friend int operator==(const XString&, const char *);    // if(x == y)
    friend int operator!=(const XString&, const char *);    // if(x != y)
    friend int operator>(const XString&, const XString&);   // if(x > y)
    friend int operator<(const XString&, const XString&);   // if(x < y)

    // case insensitive string compare
    // returns -1 if x1 < x2, 0 if x1 == x2, 1 if x1 > x2
    static int icmp(const XString& x1, const XString& x2); 

    // returns the chracter at the given position
    // or -1 if position is out of range
    int operator[](int) const;      // ch_wide = x[i];

    // returns a null terminated char * string
    char *data() const;             // char *p = x.data();
    char *c_str() const;            // char *p = x.c_str();
    operator const char*() const;   // char *p = x;

    // retruns the number of characters in the string
    int  length(void) const;        // len = x.length();
    int  size(void) const;          // len = x.length();
    bool empty(void) const;         // returns true if size()==0

    // format     
    XString &format(const char *, ...);

    // substring stuff
    XString left(int size);     // return left most sized XString
    XString right(int size);    // return right most sized XString

    // return true if string contains only whitespace characters
    bool is_white_space(void) const;   
    static bool is_white_space(const char *); 

    // return true if string contains ANY whitespace characters
    bool has_white_space(void) const;   
    static bool has_white_space(const char *); 

    // Return a new string with the leading and trailing white space removed
    XString both_trim(void) const;

    // returns true if string contains newline or linefeed characters 
    bool is_multiline(void) const;

    // find functions 
    int find(const char x, int begin = 0) const;  //returns positon in string 
    int find(const XString &x, int begin) const;         
    int find(const char *pFindStr, int begin) const;

private:

    // Allocate/Reallocate memory to hold the string.
    void adjust(int newlen, bool copyflag = false);

    // how much memory do we have allocated
    int memlen;  

    // how long is the current XString 
    int curlen;  

    // pointer to the start of memory where 
    // characters in the XString are stored
    char *str;   
};

#endif
