/* File:  TStack.cpp

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is the TStack class/library

The Initial Developer of the Original Code is Gregory Riggs.
Portions created by Gregory Riggs are Copyright (C) 2001
Gregory Riggs.  All Rights Reserved.

Contributor(s):
  Gregory Riggs <greg@nmia.com>
  ITQ LATA, LLC. <wctpxml\@itqlata.com>  

*/

#include "TStack.hpp"

/*
 *
 *    Simple pointer diagram of a Double linked list...
 *
 *    +----------------------->---------------------+
 *    |                                             |
 *    |           +------->-----+    +------->----+ |  +---+
 *    |           |             |    |            | |  |   |
 *    |  #########|#        #########|#       #########|#  |
 *    |  #*#     #*#        #*#     #*#       #*#     #*#  |
 *    |  #|#########        #|#########       #|#########  |
 *    |   |   ||             |    |            |           |
 *    +---+   |+-------<-----+    +------<-----+           |
 *            |                                            |
 *            +-------------------<------------------------+
 *
 */

//---------------------------------------------------------------------------
// TStackBase::TStackBase()
//---------------------------------------------------------------------------

TStackBase::TStackBase()
{
    Initialize();
}

//---------------------------------------------------------------------------
// TStackBase::TStackBase(const TStackBase&)
//---------------------------------------------------------------------------

TStackBase::TStackBase(const TStackBase& ts)
{
    Initialize();
    Copy(ts);
}

//---------------------------------------------------------------------------
// TStackBase::~TStackBase()
//---------------------------------------------------------------------------

TStackBase::~TStackBase()
{
    destroy(NULL);
}

//---------------------------------------------------------------------------
// TStackBase& TStackBase::operator=(const TStackBase& ts)
//---------------------------------------------------------------------------

TStackBase& 
TStackBase::operator=(const TStackBase& ts)
{
    // watch out for the special case:  ts = ts;
    if(this != &ts) 
    {
        destroy(NULL);
        Copy(ts);
    }
    return *this;
}

//---------------------------------------------------------------------------
// TStackBase::Initialize(void)
//---------------------------------------------------------------------------

void 
TStackBase::Initialize(void)
{
    sentry.next = &sentry;
    sentry.prev = &sentry;
    sentry.data = NULL;
    num = 0;
    curpos = -1;
    curptr = NULL;
}

//---------------------------------------------------------------------------
// TStackBase::Copy(const TStackBase& ts)
//---------------------------------------------------------------------------

void 
TStackBase::Copy(const TStackBase& ts)
{
    TStackNode *p;
    p = ts.sentry.next;
    while(p != &ts.sentry) 
    {
        unshift(p->data);
        p = p->next;   
    }
}

//---------------------------------------------------------------------------
// TStackNode *TStackBase::add(TStackNode *p, void *d)
//---------------------------------------------------------------------------
// This function adds a data element to the stack.
// It returns a pointer to the TStackNode used to hold the data element,
// which can be used as a key for quickly removing it form the stack
// later.

TStackNode *
TStackBase::add(TStackNode *p, void *d)
{
    TStackNode *q = new TStackNode;
    q->data = d;
    q->next = p->next;
    q->prev = p;
    p->next->prev = q;
    p->next = q;
    num++;
    curpos = -1;
    curptr = NULL;
    return q;
}

//---------------------------------------------------------------------------
// void *TStackBase::remove(TStackNode *p)
//---------------------------------------------------------------------------
// This function will remove a TStackNode form the Stack.

void *
TStackBase::remove(TStackNode *p)
{
    if(p == &sentry || p == NULL) 
    {
        return NULL;
    }

    void *d;
    d = p->data;
    p->prev->next = p->next;
    p->next->prev = p->prev;
    p->prev = NULL;  // saftey precaution in case try to remove 
    p->next = NULL;  // the same node twice, thus it more likely segfault
    p->data = NULL;  // instead of free memory read 
    delete p; 
    num--;
    curpos = -1;
    curptr = NULL;
    return d;
}

//---------------------------------------------------------------------------
// void *TStackBase::operator[](int i) 
//---------------------------------------------------------------------------
// This function allows easy access to the i_th element on the stack.
// Returns pointer to the sentry if given position is out of range.
//
// I have added add a special pointer data member to this class that 
// keeps track  of what the last accessed element was so that if elements 
// in the list are being accessed in a sequential order, the last accessed
// pointer can be used to quickly find the next node in the sequnce instead 
// of traversing down the list to get the ith element.  This allows a single 
// iteration to be O(1), but nested iterations will be O(N)

TStackNode*
TStackBase::operator[](int reqpos) 
{
    // check if requested position is within valid range
    if(reqpos < 0 || reqpos >= num) 
    {
        return &sentry;
    }

    // reset current position, if would need to traverse greater than 
    // half the length of the list to reach the requested position
    // ie.) if size is 10, user access element 1, and then accesses elment 8
    if(abs(curpos - reqpos) > num/2)
    {
        curpos = -1;
        curptr = NULL;
    }

    // if current possition is not valid, set current position to
    // either the head or tail of the list, whichever is closer
    // to the requested position
    if(curptr == NULL || curpos < 0 || curpos >= num)
    {
        if(reqpos > num/2)
        {
            curpos = num-1;
            curptr = sentry.prev;
        }
        else
        {
            curpos = 0;
            curptr = sentry.next;
        }
    }

    // walk forward through the list if needed to reach the requested position
    while(curpos < reqpos)
    {
        curptr = curptr->next;
        curpos++;
    }

    // walk backward through the list if needed to reach the requested position
    while(curpos > reqpos)
    {
        curptr = curptr->prev;
        curpos--;
    }

    // The current position should now equal the requested position
    return curptr;
}

//---------------------------------------------------------------------------
// const void *TStackBase::operator[](int i) const
//---------------------------------------------------------------------------
// A const version of operator[] method to keep the compiler happy for
// the copy constructors.  For now O(n), since we can't use the curpos 
// and curptr in a const method, the compiler complains if we do.

const TStackNode*
TStackBase::operator[](int reqpos) const
{
    TStackNode *ptr;
    int pos;

    // check if requested position is within valid range
    if(reqpos < 0 || reqpos >= num) 
    {
        return &sentry;
    }

    // set position to either the head or tail of the list, 
    // whichever is closer to the requested position
    if(reqpos > num/2)
    {
        pos = num-1;
        ptr = sentry.prev;
    }
    else
    {
        pos = 0;
        ptr = sentry.next;
    }

    // walk forward through the list if needed to reach the requested position
    while(pos < reqpos)
    {
        ptr = ptr->next;
        pos++;
    }

    // walk backward through the list if needed to reach the requested position
    while(pos > reqpos)
    {
        ptr = ptr->prev;
        pos--;
    }

    // Our position should now equal the requested position
    return ptr;
}

//---------------------------------------------------------------------------
// void TStackBase::destroy(void (*del_func)(void *))
//---------------------------------------------------------------------------
// This function takes ownership of the memory of all the data objects
// stored on the stack and deletes them.

void 
TStackBase::destroy(void (*del_func)(void *))
{
    TStackNode *p, *q;
    p = sentry.next;
    while(p != &sentry) 
    {
        q = p;
        p = p->next;
        if(del_func != NULL && q->data != NULL) 
        {
            del_func(q->data);
        }
        delete q; 
    }
    sentry.next = &sentry;
    sentry.prev = &sentry;
    num = 0;
    curpos = -1;
    curptr = NULL;
}
