/* File:  wctpXmlTransparentData.cpp

The contents of this file are subject to the Mozilla Public
License Version 1.1 (the "License"); you may not use this file
except in compliance with the License. You may obtain a copy of
the License at http://www.skytel.com/downloads/wctpxml/worldcom_mpl_1.1.txt.

Software distributed under the License is distributed on an "AS
IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
implied. See the License for the specific language governing
rights and limitations under the License.

The Original Code is the WCTP Toolkit

The Initial Developer of the Original Code is WorldCom, Inc.
Portions created by WorldCom, Inc. are Copyright  2001
WorldCom, Inc.  All Rights Reserved.

Contributor(s):
  WorldCom, Inc. <skytel-strategic-apps@wcom.com>  

*/

#include "wctpXmlTransparentData.hpp"
#include <stdio.h>

#include "EncodeUtil.hpp"
//////////////////////////////////////////////////////////////////////////////
// Constructor

wctpXmlTransparentData::wctpXmlTransparentData()
{
    Init();
}

//////////////////////////////////////////////////////////////////////////////
// Copy Constructor

wctpXmlTransparentData::wctpXmlTransparentData(
    const wctpXmlTransparentData &rval)
{
    Copy(rval);
}

//////////////////////////////////////////////////////////////////////////////
// Destructor

wctpXmlTransparentData::~wctpXmlTransparentData()
{
    Free();
}

//////////////////////////////////////////////////////////////////////////////
// Assignment Operator

wctpXmlTransparentData&
wctpXmlTransparentData::operator=(
    const wctpXmlTransparentData &rval)
{
    Free();
    Copy(rval);
    return *this;
}

//////////////////////////////////////////////////////////////////////////////
// Initialize member elements to default values

void
wctpXmlTransparentData::Init(void)
{
    wctpXmlOperation::Init();

    m_pcdata = "";

    m_encoding = WCTP_ENCODING_BASE64;
    m_type = WCTP_TRANSPARENT_OPAQUE;

    m_title = TRANSPARENT_DATA_STRING;

    m_xmlOperation = m_xmlPrefix;
    m_xmlOperation += "-";
    m_xmlOperation += m_title;
}

//////////////////////////////////////////////////////////////////////////////
// Deep copy member elements

void
wctpXmlTransparentData::Copy(const wctpXmlTransparentData &rval)
{
    //
    // Copy member variables
    //

    wctpXmlOperation::Copy(rval);

    m_encoding = rval.m_encoding;
    m_pcdata = rval.m_pcdata;
    m_type = rval.m_type;
}

//////////////////////////////////////////////////////////////////////////////
// Free any objects being held

void
wctpXmlTransparentData::Free(void)
{
    wctpXmlOperation::Free();
}

//////////////////////////////////////////////////////////////////////////////
// Clear object (reset object as if it were just constructed)

void
wctpXmlTransparentData::Clear(void)
{
    Free();
    Init();
}


//////////////////////////////////////////////////////////////////////////////
// Construct XML string for this object using given context 

XString 
wctpXmlTransparentData::AssembleOperation(XmlContext& ctx)
{
    XString dest;
    XString temp;

    temp = "<";
    temp += m_xmlOperation;
    dest += ctx.OpenIndent(temp);

    dest += ctx.XmlAttribute(
        "encoding", wctpXml::encoding_to_str(m_encoding));
    dest += ctx.XmlAttribute(
        "type", wctpXml::transparent_to_str(m_type));
    dest += ctx.IndentText(">");
    if(m_encoding == WCTP_ENCODING_BASE64)
    {
        XString encoded;
        EncodeUtil::base64_encode(encoded,m_pcdata);
        dest += ctx.XmlPcdata(encoded);
    }
    else if(m_encoding == WCTP_ENCODING_UUENCODE)
    {
        XString encoded;
        EncodeUtil::uuencode(encoded,m_pcdata);
        dest += ctx.XmlPcdata(encoded);
    }
    else
    {
        dest += ctx.XmlPcdata(m_pcdata);
    }

    temp = "</";
    temp += m_xmlOperation;
    temp += ">";
    dest += ctx.CloseIndent(temp);


    return dest;     
}


//////////////////////////////////////////////////////////////////////////////
// Parse Attribute for this object

wctp_parse_t 
wctpXmlTransparentData::ParseAttribute(
    XString& msg,
    const XString& key, 
    const XString& val)
{
    wctp_parse_t res = WCTP_PARSE_OKAY;
    if(key == "encoding")
    {
        m_encoding = wctpXml::str_to_encoding(val.c_str());
    }
    else if(key == "type")
    {
        m_type = wctpXml::str_to_transparent(val.c_str());
    }
    else
    {
        msg = "attribute |";
        msg += key;
        msg += "| does not exist for element ";
        msg += m_xmlOperation;
        res = WCTP_PARSE_ERROR;
    }
    return res;
}

//////////////////////////////////////////////////////////////////////////////
// Parse XML tree for this object 

wctp_parse_t 
wctpXmlTransparentData::ParseObject(
    XString& msg,
    XmlNode *pTree)
{
 
    //
    // make sure the element matches me
    //

    if(m_xmlOperation != pTree->element)
    {
        msg = "element mismatch, expected ";
        msg += m_xmlOperation;
        msg += ", but got |";
        msg += pTree->element;
        msg += "|";
        return WCTP_PARSE_ERROR;  
    }


    //
    // load in attribute values
    //

    int i;
    for(i=0; i<pTree->attributes.size(); i++)
    {
        XmlAttribute *pAttribute = pTree->attributes[i];
        wctp_parse_t res = ParseAttribute(
            msg, pAttribute->key, pAttribute->val);
        if(res != WCTP_PARSE_OKAY)
        {
            return res;
        }
    }

    //
    // collect pcdata 
    //

    if(m_encoding == WCTP_ENCODING_BASE64)
    {
        XString decoded;
        bool flag = EncodeUtil::base64_decode(decoded, pTree->pcdata);
        if(!flag)
        {
            msg = "Error decoding base64 data";
            return WCTP_PARSE_ERROR;
        }
        m_pcdata = decoded;
    }
    else if(m_encoding == WCTP_ENCODING_UUENCODE)
    {
        XString decoded;
        bool flag = EncodeUtil::uudecode(decoded, pTree->pcdata);
        if(!flag)
        {
            msg = "Error decoding uuencoded data";
            return WCTP_PARSE_ERROR;
        }
        m_pcdata = decoded;
    }
    else
    {
        m_pcdata = pTree->pcdata;
    } 
    EncodeUtil::hexdump(m_pcdata_hexdump,m_pcdata); 

    //
    // Should not be any child objects
    //

    if(pTree->children.size() > 0)
    {
        msg = m_xmlOperation;
        msg += " element does not accept any child elements";
        return WCTP_PARSE_ERROR;
    }
    
    return WCTP_PARSE_OKAY;
}

//////////////////////////////////////////////////////////////////////////////
// Access operation for m_encoding attribute

wctp_encoding_t 
wctpXmlTransparentData::GetEncoding(void) const 
{
    return m_encoding;
}

//////////////////////////////////////////////////////////////////////////////
// Settor operation for m_encoding attribute

void  
wctpXmlTransparentData::SetEncoding(wctp_encoding_t newValue)
{
    m_encoding = newValue;
}

//////////////////////////////////////////////////////////////////////////////
// Access operation for m_pcdata attribute

const char * 
wctpXmlTransparentData::GetPcdata(void) const 
{
    return m_pcdata;
}

//////////////////////////////////////////////////////////////////////////////
// Settor operation for m_pcdata attribute

void  
wctpXmlTransparentData::SetPcdata(const char * newValue)
{
    m_pcdata = newValue;
}

//////////////////////////////////////////////////////////////////////////////
// 

int
wctpXmlTransparentData::GetPcdataSize(void) const
{
    return m_pcdata.size();
}

//////////////////////////////////////////////////////////////////////////////
// 

void
wctpXmlTransparentData::SetPcdata(const char *data, int len) 
{
    EncodeUtil::hexdump(m_pcdata_hexdump,m_pcdata); 
    m_pcdata = XString(data, len);
}

//////////////////////////////////////////////////////////////////////////////
// 

const char *
wctpXmlTransparentData::GetPcdataHexdump(void) const
{
    return m_pcdata_hexdump;
}

//////////////////////////////////////////////////////////////////////////////
// Access operation for m_type attribute

wctp_transparent_t 
wctpXmlTransparentData::GetType(void) const 
{
    return m_type;
}

//////////////////////////////////////////////////////////////////////////////
// Settor operation for m_type attribute

void  
wctpXmlTransparentData::SetType(wctp_transparent_t newValue)
{
    m_type = newValue;
}

